#!/usr/bin/env perl

# This CGI script parses the PATH_INFO variable for bounding
# coordinates and generates a map of the area of interest (AOI)
# that has a width specified in the QUERY_STRING variable.
# The projection used is Plate Carree.
#
# Norman Kuring		18-Nov-1999
# Norman Kuring		17-Sep-2002

# Programs executed by this script.
use OBPG::Launder qw(launder);
our $ppmtogif   = 'ppmtogif -quiet';
our $pscoast    = "gmt pscoast";
our $gmtset     = "gmt gmtset " . join ' ', qw@
                 ANOT_FONT_SIZE 8
                 BASEMAP_TYPE plain
                 MEASURE_UNIT inch
                 FRAME_WIDTH 0.06
                 TICK_LENGTH 0.1
                 PAGE_COLOR 255/255/255
                 PLOT_DEGREE_FORMAT dddF
                 @;
our $gs         = 'gs -sDEVICE=pnmraw -sOutputFile=- -q -r100';

our ($n,$s,$w,$e,$width);
if(defined $ENV{PATH_INFO}){

  # To avoid confusing a potential user, require that the
  # PATH_INFO variable -- if specified -- end in "gif",
  # since this script produces images in the GIF format.
  $ENV{PATH_INFO} =~ /gif$/i or die qq@I won't play nice unless you end the @,
                                    qq@PATH_INFO part of your URL in "gif". @,
                                    qq@Died@;

  ($n,$s,$w,$e) = string_to_bounds($ENV{PATH_INFO});
  if(defined $ENV{QUERY_STRING} and $ENV{QUERY_STRING} > 30){
    $width = launder( $ENV{QUERY_STRING} );
  }
  else{
    $width = 500;
  }
}
else{
  ($n,$s,$w,$e) = (90,-90,-180,180);
  $width = 500;
}

# Determine the scale.  A scale of 1 represents one pixel per 0.01 degrees.
our $degrees_wide = deltalon($w,$e);
$degrees_wide > 0 or
  die "The western and eastern boundaries of the AOI must not be the same.";
our $scale = $width/$degrees_wide/100;

# Pick a GMT database appropriate to the scale.
if($scale > 5){
  $gmtD = "-Df";      # full resolution
}
elsif($scale > 0.5){
  $gmtD = "-Dh";      # high resolution
}
elsif($scale > 0.1){
  $gmtD = "-Di";      # intermediate resolution
}
elsif($scale > 0.05){
  $gmtD = "-Dl";      # low resolution
}
else{
  $gmtD = "-Dc";      # crude resolution
}

# Get rid of any dateline discontinuity in the bounding longitudes.
# GMT doesn't like those.
$e += 360 if $w > $e;

# Compute the center longitude.
our $clon = ($w + $e)/2;

# Compute the height of the map.
our $height = int(($n - $s) * 100 * $scale + 0.5);

our $gmtw = $width/100;

# Choose an annotation interval that is an even multiple of a minute of arc.
# The bitmap representations of the latitudes or longitudes get to be too
# wide if they must include degrees, minutes, *and* seconds.
our $ann = int(60/$scale) || 1;
our $tick = $ann/10;
$width += 130;
$height += 70;

# Compose and run the command that generates the map, and send
# the resulting image through a format converter to standard
# output with an HTTP header prepended.
our $command = "$gmtset; $pscoast -R$w/$e/$s/$n -JQ$clon/$gmtw " .
              "-P -X0.65 -Y0.35 -B${ann}mf${tick}m -W -N1 -N2 $gmtD |" .
              "$gs -g${width}x$height - | $ppmtogif |";

# The following two lines prevent pscoast from writing warnings to
# standard error.
our $tmpdir = "/tmpfs01/aoimap.$^T.$$";
mkdir $tmpdir,0755;
#$ENV{HOME} = $tmpdir;	# Prevents warning:
			# "GMT Warning: Could not determine home directory!"
chdir $tmpdir;		# Prevents warning:
			# "GMT Fatal Error: Could not update .gmtcommands file"

open P,$command or die qq@Error opening pipe from, "$command". Died@;
local($/) = undef;
our $gif = <P>;
close P or die qq@Error executing pipeline, "$command". Died@;
#chdir;
system("rm -rf $tmpdir");
our $len = length $gif;
print "Content-type: image/gif\nContent-length: $len\n\n$gif";



# Parse geographical boundaries from a specially-formatted string.
sub string_to_bounds{
  my $string = shift;
  my ($n,$nh,$s,$sh,$w,$wh,$e,$eh) = $string =~
     /(?:[^\d]|\d{4,})
     \.(\d+(?:\.\d+)?)([NS])
     \.(\d+(?:\.\d+)?)([NS])
     \.(\d+(?:\.\d+)?)([EW])
     \.(\d+(?:\.\d+)?)([EW])
     /x;
  defined $n or return ();
  $n = $nh eq "N" ? $n : -$n;
  $s = $sh eq "N" ? $s : -$s;
  $w = $wh eq "E" ? $w : -$w;
  $e = $eh eq "E" ? $e : -$e;
  $n > 90 || $n < -90 || $s > 90 || $s < -90 and return ();
  $w += 360 while($w < -180);
  $w -= 360 while($w >  180);
  $e += 360 while($e < -180);
  $e -= 360 while($e >  180);
  ($n,$s,$w,$e);
}

sub bounds_to_string{
  my ($n,$s,$w,$e) = @_;
  $w += 360 while($w < -180);
  $w -= 360 while($w >  180);
  $e += 360 while($e < -180);
  $e -= 360 while($e >  180);
  my ($nh,$sh,$wh,$eh);
  ($n,$nh) = $n < 0 ? (-$n, 'S') : ($n, 'N');
  ($s,$sh) = $s < 0 ? (-$s, 'S') : ($s, 'N');
  ($w,$wh) = $w < 0 ? (-$w, 'W') : ($w, 'E');
  ($e,$eh) = $e < 0 ? (-$e, 'W') : ($e, 'E');
  "$n$nh.$s$sh.$w$wh.$e$eh";
}

sub deltalon{
  my ($west,$east) = @_;
  $west += 360 while $west < -180;
  $west -= 360 while $west >  180;
  $east += 360 while $east < -180;
  $east -= 360 while $east >  180;
  $west > $east ? 360 + $east - $west : $east - $west;
}
