#!/usr/bin/env perl
# Norman Kuring		18-May-2000

# Norman Kuring		 6-Jun-2001 Fixed imagemap for MS Internet Explorer
# Norman Kuring		 8-Jan-2004 Fixed bug in getting dimensions of
#				    coverage map image.
# Norman Kuring		 9-Jan-2004 Chdir to /tmp instead of $tmpdir to fix
#				    a bug discovered on webserver1 where GMT
#				    could not write its .gmtcommands file to
#				    $tmpdir
# Norman Kuring		23-Feb-2004 Remove trailing whitespace from StationInfo
#				    lines.  A space after the station code
#				    prevented that station's data from being
#				    properly displayed.
# Norman Kuring		 6-Dec-2004 New webserver (Sun) and new directory
#				    hierarchy
# Norman Kuring		 2-Dec-2005 Add hyperlink to our ground station webcam
# Norman Kuring		 7-Mar-2006 Change browse.pl?typ=x from HRPT to MLAC
# Norman Kuring		15-Jan-2014 The SeaWiFS mission ended in 2010.
#				    No need to extend the graph further.
# Norman Kuring		17-Jan-2014 Sort stations labeled with "no data" to the
#				    end of the list in the default "code" sort.
# Norman Kuring		 1-Sep-2021 Hardcode the path to the stafilgraph
#				    program since the default version was
#				    not getting the correct shared libraries.

use CGI 'header';
use OBPG::Launder qw(launder);

#$ENV{GS_DEVICE} = "ppmraw";
#$ENV{GS_OPTIONS} = "-dNOPAUSE -q  -r100 -sOutputFile=-";

our $station_info_file = '/web/oceancolor/metadata/StationInfo';
our $brsdir = '/web/data/SeaWiFS/L1BRS/HRPT';

our $tmpdir = "/tmpfs01";

our $thisScript = '/cgi/' . (split /\//,$0)[-1];
our $browse = '/cgi/browse.pl';
our $hrpt_received = '/cgi/hrpt_received.pl';

our $pi = 3.14159265358979323846;

our $search_radius = 500;	# kilometers
our $location_map_width = 300;
our $location_map_height = int($location_map_width/2);

use IO::Handle;
use IPC::Run qw(run start finish);

our $pnmtopng = "pnmtopng";
our $pnmquant = "pnmquant 256";
our $identify = "identify";
our $gs = "gs";
our $pscoast = "gmt pscoast -P -X0 -Y0 -Dc";
our $psxy = "gmt psxy -P -X0 -Y0";

# Display the StationInfo file without any HTML tags if so requested.
if($ENV{QUERY_STRING} =~ /^PLAIN$/i){
  &plaintext;
  exit;
}

# Parse the StationInfo file.
our $station_info = &read_station_info;

use OBPG::CGI qw(ReadParse);
our %in = ReadParse();

# Filter out shell metacharacters
foreach my $key ( keys %in ){
  $in{$key} = launder( $in{$key}, "\$in{$key}" );
}

# If the user clicks on the map of station locations, I search
# the station_info for all stations within $search_radius kilometers
# of the click and then construct a list of those station codes
# and place it in $in{S}.  If no stations are found within the
# $search_radius then I send a "No Response" header to the client
# browser.
if(defined $in{'M.x'}){
  my $width = $in{'M.w'} || $location_map_width;
  my $lat = (90 - $in{'M.y'} * 360 / $width) * $pi/180;
  my $lon = ($in{'M.x'} * 360 / $width - 180) * $pi/180;
  my @s;
  foreach (sort keys %$station_info){
    defined $station_info->{$_}{lat} and
    defined $station_info->{$_}{lon} and
    gcdist($lat,$lon,
           $station_info->{$_}{lat} * $pi/180,
           $station_info->{$_}{lon} * $pi/180  ) < $search_radius
    and push @s,$_;
  }
  if(@s == 0){
    print header(-status=>204);
    return;
  }
  $in{S} = join '@',@s;
}

if($ENV{PATH_INFO} and $ENV{PATH_INFO} =~ /\.png$/i){
  if($in{S}){
    my @stations = split /@/, $in{S};
    send_locationmap(@stations);
  }
  else{
    send_stafilgraph($ENV{PATH_INFO});
  }
}
elsif(defined $in{S}){
  my @stations = split /@/, $in{S};
  if(@stations == 1){
    show_station($in{S});
  }
  else{
    show_list(@stations);
  }
}
else{
  show_list();
}
exit;


sub send_stafilgraph{
  my $pathinfo = shift;
  my $stafilgraph = '/web/oceancolor/bin/stafilgraph';
  my ($station) = $pathinfo =~ /^\/([a-z]{3})\./i;
  defined $station or $station = "undefined_code";
  my $pipe = "$stafilgraph $station | $pnmtopng -quiet |";
  open P, $pipe or die qq@Could not open pipe from, "$pipe". ($!) Died@;
  local($/) = undef;
  my $png = <P>;
  close P or die qq@Error executing pipe from, "$pipe". ($!) Died@;
  my $len = length $png;
  print "Content-type: image/png\nContent-length: $len\n\n$png";
}

sub send_locationmap{
  my @stations = @_;
  @stations or @stations = keys %$station_info;

  # Go to a directory where GMT may write its .gmtcommands
  # and .gmtdefaults files.  Also make that the HOME directory
  # to avoid another GMT complaint.
  chdir '/tmp';
  $ENV{HOME} = $tmpdir;

  # Generate the background map in PostScript format.
  my $hectowidth = $location_map_width/100;
  my $pipe =
  "$pscoast -Rd -JQ0/$hectowidth -G51/102/51 -S51/51/102 -K|";
  open P, $pipe or die qq@Could not open pipe from "$pipe". Died@;
  local($/) = undef;
  my $ps = <P>;
  close P or die qq@Error executing pipe from "$pipe". Died@;
  
  # Generate the overlay showing station locations.
  my $npsxy  = "gmt psxy";
  my @cmd = ("$npsxy","-P","-X0","-Y0","-Rd","-JQ0/$hectowidth","-G255/0/0","-Sc0.05","-O");
  my $h = start(\@cmd, '<pipe', \*WFH, '>pipe', \*RFH)
           or die "Error opening 2-way pipeline to @cmd ($!)" ;
  foreach (@stations){
    if(defined $station_info->{$_}{lat} and defined $station_info->{$_}{lon}){
      print WFH "$station_info->{$_}{lon} $station_info->{$_}{lat}\n";
    }
  }
  close WFH;
  $ps .= <RFH>;
  close RFH;
  $h->finish;

  # Convert the PostScript into a PNG image.
  my @cmd = ($gs,"-sDEVICE=pngalpha","-sOutputFile=-","-q","-r100","-g${location_map_width}x$location_map_height","-");
  my $png;
  run \@cmd,\$ps,\$png;
  my $len = length $png;
  print "Content-type: image/png\nContent-length: $len\n\n$png";
}

sub show_list{
  my @stalist = @_;
  my $total_stations = keys %$station_info;

  # Allow the user to select all stations that have no coordinates specified.
  my @mobile_stations;
  foreach (sort keys %$station_info){
    defined $station_info->{$_}{lat} and
    defined $station_info->{$_}{lon} or
    push @mobile_stations,$_;
  }
  my $mobile_stations = '';
  if(@mobile_stations > 0){
    $mobile_stations = join '@',@mobile_stations;
    $mobile_stations =
    qq@, or you may <a href="$thisScript?S=$mobile_stations">select\n@ .
    qq@all stations that have no fixed location</a>@;
  }

  my $s = "";
  my $stalist = "";
  if(@stalist){
    # Remove all nonspecified stations from the station_info hash.
    my %s;
    grep ++$s{$_} && 0, @stalist;
    grep {$s{$_} || delete $station_info->{$_} and 0} keys %$station_info;

    # Arrange to pass along the selected list of stations.
    $stalist = join '@', sort @stalist;
    $s = "&S=$stalist";
  }
  my $sort_field = defined $in{F} ? lc($in{F}) : 'code';
  my @stations = sort_stations($sort_field);
  my $numstations = @stations;
  my $display_field = $sort_field eq 'code' ? 'station' : $sort_field;
  my @sortable = qw/
                 code
                 name
                 e-mail
                 station
                 datacen
                 descrip
                 affil
                 status
                 /;
  my $sortcontrol =
  qq@<table border><tr><th align="left">@ .
  qq@Stations are shown sorted by the <i>\u$sort_field</i> field.\n@ .
  qq@They can also be sorted by the folowing fields.\n@ .
  qq@</tr>\n@;
  foreach (@sortable){
    next if $_ eq $sort_field;
    $sortcontrol .= qq@<tr><TD align="center">@ .
                    qq@<a href="$thisScript?F=$_$s">\u$_</a></td></tr>\n@;
  }
  $sortcontrol .= qq@</table>\n@;

  my $heading;
  if($numstations < $total_stations){
    $heading =
    qq@$numstations of <a href="$thisScript">$total_stations</a> @ .
    qq@HRPT Stations@;
  }
  else{
    $heading = "$total_stations HRPT Stations";
  }

  my $locimg =
  qq@You may also choose stations by clicking on the map below@ .
  $mobile_stations .
  qq@.\nAll fixed stations within $search_radius\n@ .
  qq@kilometers of a click on the map will be selected.\n@ .
  qq@<form method="post" action="$thisScript">\n@ .
  qq@<input name="M" type="image"\n@ .
  qq@src="$thisScript/HRPT_locations.png?S=$stalist" alt="coverage map"\n@ .
  qq@width="$location_map_width" height="$location_map_height">\n</form>@;

  # Find out when the StationInfo file was last updated.
  my @wday = qw/Sun Mon Tue Wed Thu Fri Sat/;
  my @mon = qw/Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec/;
  my $last_update = $^T - int((-M $station_info_file)*86400);
  my @t = gmtime($last_update);
  $last_update =
  sprintf "$wday[$t[6]], $t[3] $mon[$t[4]] %d at %02d:%02d:%02d UT",
  $t[5] + 1900, $t[2], $t[1], $t[0];

  head('SeaWiFS HRPT Stations');
  print
  qq@<h1>\n@,
  qq@$heading\n@,
  qq@<i>\n@,
  qq@<font size="-1">(Station list last updated on $last_update.)@,
  qq@ <a href="/seawifs/nsg_webcam/lastsnap.jpg">status</a></font>\n@,
  qq@</i>\n@,
  qq@<br />\n@,
  qq@<font size="-1">A <a href="$thisScript?PLAIN">plain text version</a>\n@,
  qq@of this information is also available.</font>\n@,
  qq@</h1>\n@,
  qq@<table><tr valign="top"><td>\n@,
  qq@<table border cellspacing="0" cellpadding="0">\n@,
  qq@<tr align="left"><th>Code</th><th>\u$display_field</th></tr>\n@;
  foreach (@stations){
    my $val;
    if(defined $station_info->{$_}{$display_field}){
      $val = $station_info->{$_}{$display_field};
      $val =~ s/\n/<br \/>\n/g;
    }
    else{
      $val = qq@<i>not available</i>@;
    }
    print qq@<tr><td><a href="$thisScript?S=$_">$_</a></td>@,
          qq@<td>$val</td></tr>\n@;
  }
  print qq@</table>\n</td><td>\n@,
        qq@<table><tr><td>\n$sortcontrol</td></tr>\n@,
        qq@<tr><th>\n$locimg\n</th></tr></table>\n</td></tr></table>\n@;
  &tail;
}

sub sort_stations{
  my $sort_field = shift;
  my @s = keys %$station_info;
  if($sort_field =~ /^name$/){
    @s = sort {
         my $aa = $station_info->{$a}{name} || ""; # Get the "name" field
         my $bb = $station_info->{$b}{name} || ""; # from this station's info.
         $aa =~ s/\(.*\)//g;	# Get rid of parenthesized
         $bb =~ s/\(.*\)//g;	# expressions e.g. (President).
         $aa =~ s/(jr|sr)\s*$//im;	# Remove any "Jr", or "Sr" from
         $bb =~ s/(jr|sr)\s*$//im;	# the end of the first line.
         ($aa) = $aa =~ /(\S*)\s*$/m;	# Select the last word of the 1st line.
         ($bb) = $bb =~ /(\S*)\s*$/m;	# I will assume this is a surname.
         $aa =~ s/\W//g;	# Remove non-alpha-numerics
         $bb =~ s/\W//g;	# before comparing.
         lc($aa) cmp lc($bb);	# Case-insensitive comparison
         } @s;
  }
  elsif($sort_field =~ /^e-mail$/){
    @s = sort {
         my $da = defined $station_info->{$a}{'e-mail'};
         my $db = defined $station_info->{$b}{'e-mail'};
         $da or $db or return 0;
         $da and !$db and return 1;
         $db and !$da and return -1;
         # Split email addresses into a name and a host.
         my ($an,$ah) = split /\@/, $station_info->{$a}{'e-mail'};
         my ($bn,$bh) = split /\@/, $station_info->{$b}{'e-mail'};
         $ah = (split ' ',$ah)[0];	# Remove extraneous text
         $bh = (split ' ',$bh)[0];	# from host.
         my @ah = split /\./,$ah;	# Split hosts into individual
         my @bh = split /\./,$bh;	# components (e.g. nasa and gov)
         @ah = reverse @ah;
         @bh = reverse @bh;
         push @ah,$an;
         push @bh,$bn;
         my $c;
         my ($aa,$bb) = ("","");
         while(defined $aa and defined $bb){
           $aa = shift @ah;
           $bb = shift @bh;
           $c = lc($aa) cmp lc($bb);
           $c and last;
         }
         $c;
         } @s;
  }
  elsif($sort_field =~ /^code$/){
    @s = sort {
         my $isa = $station_info->{$a}{station} !~ /no data/i ? 1 : 0;
         my $isb = $station_info->{$b}{station} !~ /no data/i ? 1 : 0;
         $isb <=> $isa or
         $a cmp $b
    } @s;
  }
  else{
    @s = sort {
         lc($station_info->{$a}{$sort_field}) cmp
         lc($station_info->{$b}{$sort_field}) or
         $a cmp $b
         } @s;
  }
  @s;
}

sub show_station{
  my $sta = shift;
  # The following list indicates the order in which the standard
  # fields should be displayed.
  my @fields = qw/
               name
               affil
               address
               phone
               fax
               e-mail
               url
               lat
               lon
               vdate
               status
               datacen
               station
               descrip
               /;
  my $descr = $station_info->{$sta}{station} || $sta;

  # Find out how many browse files we have for this station.
  my $scene_count = count_scenes($sta);
  my $scenes_available;
  if($scene_count == 0){
    $scenes_available = "No scenes have been received from this station.";
  }
  elsif($scene_count == 1){
    $scenes_available =
    qq@<a href="$hrpt_received?$sta">One scene</a>\n@ .
    qq@is available from this station.@;
  }
  else{
    $scenes_available =
    qq@<a href="$hrpt_received?$sta">$scene_count scenes</a>\n@ .
    qq@are available from this station.@;
  }

  head("SeaWiFS HRPT Station : $descr");
  print
  qq@<table>\n@,
  qq@<tr>\n@,
  qq@<td><h1>$descr ($sta)</h1></td>\n@,
  qq@<td>$scenes_available</td>\n@,
  qq@</tr><tr>\n<td>\n@,
  qq@<table border>\n@,
  qq@<tr><th colspan="2">Standard Information</th></tr>\n@;
  foreach (@fields){
    my $val = "<i>not available</i>";
    if(defined $station_info->{$sta}{$_}){
      $val = $station_info->{$sta}{$_};
      /^url/ and $val =~ s@(\S+)@<a href="$1" target="_blank">$1</a>@g;
      /^e-mail/ and $val =~ s#([\w.\-]+\@[\w.\-]+)#<a href="mailto:$1">$1</a>#g;
      $val =~ s/\n/<br \/>\n/g;
    }
    my $key = ucfirst $_;
    print qq@<tr>\n<td>\n$key\n</td>\n<td>\n$val\n</td>\n</tr>\n@;
    delete $station_info->{$sta}{$_};
  }
  my @keys = sort keys %{ $station_info->{$sta} };
  if(@keys){
    print qq@<tr><th colspan="2"><font size="-2">@,
          qq@Additional Fields</font></th></tr>\n@;
    foreach (@keys){
      my $val = $station_info->{$sta}{$_};
      /^url/ and $val =~ s@(\S+)@<a href="$1" target="_blank">$1</a>@g;
      /^e-mail/ and $val =~ s#(\S+)#<a href="mailto:$1">$1</a>#g;
      $val =~ s/\n/<br \/>\n/g;
      my $key = ucfirst $_;
      $key = qq@<font size="-2">$key</font>@;
      $val = qq@<font size="-2">$val</font>@;
      print qq@<tr>\n<td>\n$key\n</td>\n<td>\n$val\n</td>\n</tr>\n@;
    }
  }
  print qq@</table>\n</td>\n@;

  # Display a coverage map if one exists.
  my $coverage_map;
  # if(-s "$ENV{DOCUMENT_ROOT}SeaWiFS/HRPT/MAPS/$sta.gif"){
  if(-s "/data1/www-data/oceancolor.gsfc.nasa.gov/htdocs/seawifs/hrpt/maps/$sta.png"){
    my $dimensions =
    (split ' ',`$identify /data1/www-data/oceancolor.gsfc.nasa.gov/htdocs/seawifs/hrpt/maps/$sta.png`)[2];
    my ($w,$h) = $dimensions =~ /^(\d+)x(\d+)/;
    $coverage_map = qq@<img src="/seawifs/hrpt/maps/$sta.png" @ .
                    qq@width="$w" height="$h" alt="$sta coverage map">@;
  }
  elsif(-s "/data1/www-data/oceancolor.gsfc.nasa.gov/htdocs/seawifs/hrpt/maps/$sta.gif"){
    my $dimensions =
    (split ' ',`$identify /data1/www-data/oceancolor.gsfc.nasa.gov/htdocs/seawifs/hrpt/maps/$sta.gif`)[2];
    #(split ' ',`$identify $ENV{DOCUMENT_ROOT}SeaWiFS/HRPT/MAPS/$sta.gif`)[2];
    my ($w,$h) = $dimensions =~ /^(\d+)x(\d+)/;
    $coverage_map = qq@<img src="/seawifs/hrpt/maps/$sta.gif" @ .
                    qq@width="$w" height="$h" alt="$sta coverage map">@;
  }
  else{
    $coverage_map = "There is no coverage map available for this station.";
  }
  print qq@<td>$coverage_map</td>\n</tr>\n@;

  # Display a graph of file distribution if we
  # have received any files from this station.
  my ($filegraph, $imgmap);
  # The height of the station file graph depends on the number of
  # years represented.  That graph -- as drawn by the stafilgraph
  # program -- should cover all the years from 1997 to 2010.
  my $numyears = 14;
  my $filegraph_height = 32 + $numyears * 34;
  my $filegraph_width = 420;
  if($scene_count > 0){
    my $imgmapname = "stafilgraph";
    $filegraph =
    qq@<img src="$thisScript/$sta.file_graph.png" usemap="#$imgmapname"\n@ .
    qq@alt="graph showing for which days we have received data from the@ .
    qq@ $sta station" width="$filegraph_width" height="$filegraph_height">\n@ .
    qq@<br />\n<font size="-1">\n@ .
    qq@Click on any portion of the timeline to go to the HRPT browser\n@ .
    qq@for this station on that day.\n</font>\n@;
    $imgmap = imgmap($imgmapname,$sta);
  }
  else{
    $filegraph = 
    qq@<img src="$thisScript/$sta.file_graph.png"\n@ .
    qq@alt="graph showing that we have not received any files from the@ .
    qq@ $sta station" width="$filegraph_width" height="$filegraph_height">\n@;
    $imgmap = "";
  }
  print
  qq@<tr>\n@,
  qq@<td>$filegraph</td>\n@,
  qq@<td>\n@,
  qq@<font size="-1">\n@,
  qq@The graph on the left shows the days on which this station\n@,
  qq@collected transmissions from the Orbview 2 satellite.  Each\n@,
  qq@vertical line along the timeline represents a single day, and\n@,
  qq@the height of that line indicates how many separate scenes were\n@,
  qq@collected on that day <i>(1, 2, 3, 4, or more than 4)</i>.  Most\n@,
  qq@stations receive one to three passes per day.  High-latitude\n@,
  qq@stations get more because the satellite flies within range of\n@,
  qq@their antennas more often during a day. <i>(If a lower-latitude\n@,
  qq@station gets more than three passes per day, that may be\n@,
  qq@an indication of some problem in the ground processing of\n@,
  qq@the satellite data stream.)</i>\n@,
  qq@<p>\n@,
  qq@The color-coding of the vertical lines gives some indication of\n@,
  qq@how long ago the SeaWiFS Project received the data from the remote\n@,
  qq@station.  <i>(These colors are actually based on the creation dates\n@,
  qq@of level-1 browse files on our system, but, since we generally\n@,
  qq@create browse files shortly after receiving the data, the meaning\n@,
  qq@is essentially the same.)</i>\n@,
  qq@</font>\n@,
  qq@</td>\n@,
  qq@<tr>\n@,
  qq@</table>\n@,
  qq@$imgmap@;
  &tail;
}

sub count_scenes{
  my $station = shift;
  opendir D,$brsdir or die qq@Could not open directory, "$brsdir". ($!) Died@;
  my @ydirs = grep /^\d{4}$/, readdir D;
  closedir D;
  my $numfiles = 0;
  my $d0;
  foreach $d0 (@ydirs){
    substr($d0,0,0) = "$brsdir/";
    opendir D,$d0 or die qq@Could not open directory, "$d0". ($!) Died@;
    my @ddirs = grep /^\d{3}$/, readdir D;
    closedir D;
    my $d1;
    foreach $d1 (@ddirs){
      substr($d1,0,0) = "$d0/";
      opendir D,$d1 or die qq@Could not open directory, "$d1". ($!) Died@;
      $numfiles += grep /^S\d{13}\.L1A_H${station}_BRS$/, readdir D;
      closedir D;
    }
  }
  $numfiles;
}

# Compute the great-circle distance between two earth locations.
# Latitudes and longitudes should be expressed in radians.
# The distance returned is specified in kilometers.
#
# See:
#   Map Projections Used by the U.S. Geological Survey
#   Geological Survey Bulletin 1532
#   by John P. Snyder
#   1983.  2nd ed.   p. 34.
sub gcdist{
  my ($lat1,$lon1,$lat2,$lon2) = @_;
  my $meanearthradius = 6371.229;	# kilometers
  my $sinlat1 = sin($lat1);
  my $coslat1 = cos($lat1);
  my $cos2lat1 = $coslat1 * $coslat1;
  my $sinlat2 = sin($lat2);
  my $coslat2 = cos($lat2);
  my $deltalon = $lon1 - $lon2;
  my $sindelta = sin($deltalon);
  my $cosdelta = cos($deltalon);
  my $cosdist = $sinlat2*$sinlat1 + $coslat2*$coslat1*$cosdelta;
  my $stuff = $coslat2*$sinlat1 - $sinlat2*$coslat1*$cosdelta;
  my $sindist = sqrt($cos2lat1*$sindelta*$sindelta + $stuff*$stuff);
  my $distance = abs(atan2($sindist, $cosdist)) * $meanearthradius;
}

sub imgmap{
  # The dimensions used in this subroutine are intended to match
  # those used in the stafilgraph program.
  my ($name,$station) = @_;
  my $today = int($^T/86400);
  my $firstday = 10108;		# 4-Sep-1997
  my $html = qq@<map name="$name">\n@;
  my $day;
  for($day = $firstday; $day <= $today; $day++){
    my ($year,$dayofyear) = dayofyear($day);
    my $x1 = 4 + $dayofyear - 1;
    my $y1 = 4 + ($year - 1997) * 34;

			# Imagemap will not work for MSIE without the "+ 1"
    my $x2 = $x1 + 1;	# here.  Apparently Microsoft interprets rectangle
			# boundaries differently from Netscape.

    my $y2 = $y1 + 33;
    my $datestr = datestr($day);
    $html .= qq@<area shape="rect" coords="$x1,$y1,$x2,$y2"@ .
             qq@ alt="$datestr"@ .
             qq@ href="$browse?per=DAY&day=$day&typ=MLAC">\n@;
  }
  $html .= qq@</map>\n@;
}

sub dayofyear{
  my $day = shift;
  my ($year,$dayofyear) = (gmtime $day * 86400)[5,7];
  ($year + 1900, $dayofyear + 1);
}

sub datestr{
  my $day = shift;
  my ($d,$m,$y) = (gmtime $day*86400)[3,4,5];
  my @mon = qw/Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec/;
  sprintf "%02d$mon[$m]%d",$d,$y+1900;
}


sub head{
  my $title = shift;
  print
  qq@Content-type: text/html\n\n@,
  qq@<html lang="en">\n@,
  qq@<head>\n@,
  qq@<title>$title</title>\n@,
  qq@</head>\n@,
  qq@<body bgcolor="#FFFFFF" text="#000000">\n@,
  qq@<center><h1>SeaWiFS Authorized HRPT Station List</h1></center>\n@,
##  qq@This is the current list of the authorized NASA research\n@,
##  qq@stations for SeaWiFS high resolution LAC data.\n@,
##  qq@These are stations which have submitted the\n@,
##  qq@<a href="/SeaWiFS/ANNOUNCEMENTS/APPENDIX_C.html">Letter Agreement</a>\n@,
##  qq@for SeaWiFS Direct Readout Ground Stations from Appendix C of the\n@,
##  qq@<a href="/SeaWiFS/ANNOUNCEMENTS/DEAR_COLLEAGUE.html">"Dear\n@,
##  qq@Colleague"</a> solicitation or which are operated by GeoEye.\n@,
##  qq@<p>\n@,
##  qq@We will update this list as new agreements are received;\n@,
##  qq@there is no deadline.  If you wish to get SeaWiFS data or to\n@,
##  qq@establish a station, please submit the forms; remember, we\n@,
##  qq@can't give you SeaWiFS data without them, and researchers\n@,
##  qq@who do get the data from us can only share them with people\n@,
##  qq@on the list.\n@,
##  qq@<p>\n@,
##  qq@<b>Note:</B> Only <i>operational</i> satellite receiving\n@,
##  qq@stations will be provided with decryption boxes by\n@,
##  qq@<a href="http://www.geoeye.com" target="_blank">GeoEye</a>.\n@,
##  qq@Please verify that you are designated as <i>operational</i>\n@,
##  qq@on this list.  If you are incorrectly listed, please notify\n@,
##  qq#<a href="mailto:gene\@seawifs.gsfc.nasa.gov">Gene Feldman</a> of\n#,
##  qq@your current status.\n@,
##  qq@<p>\n@,
##  qq@<hr />\n@,
##  qq@<b>\n@,
##  qq@Follow this <a href="/SeaWiFS/LICENSE/checklist.html"\n@,
##  qq@target="_blank">CHECKLIST</a> to either become an\n@,
##  qq@<i>Authorized SeaWiFS Data User</i> or an <i>Authorized\n@,
##  qq@SeaWiFS Direct Readout Ground Station</i>\n@,
  qq@</b>\n@,
  qq@<hr />\n@;
}

sub tail{
  print
  qq@<p>\n@,
  qq@<center>\n@,
  qq@<a href="/"><img src="/icons/oceancolor_icon.png"\n@,
  qq@width="371" height="50" border="0" alt="OceanColor Web"></a>\n@,
  qq@<hr />\n@,
  qq@<address>\n@,
  qq@<a href="/staff/gene/">gene carl feldman</a>\n@,
  qq#(<a href="mailto:gene.c.feldman\@nasa.gov">#,
  qq#gene.c.feldman\@nasa.gov</a>)\n#,
  qq@(301) 286-9428\n@,
  qq@</address>\n@,
  qq@</center>\n@,
  qq@</body>\n@,
  qq@</html>\n@;
}

sub plaintext{
  open F,'<',$station_info_file or die qq@Could not open file, @,
                                   qq@"$station_info_file". ($!) Died@;
  local($/) = undef;
  my $text = <F>;
  close F;
  print "Content-type: text/plain\n\n$text";
}

sub read_station_info{
  # This function reads HRPT station information into a hash of hashes.
  # The top-level hash is keyed by three-letter station codes.
  # The second-tier hashes are keyed by strings like "lat", "lon",
  # "url", etc.
  open F,'<', $station_info_file or
    die qq@Could not open file, "$station_info_file". ($!) Died@;
  local($/) = undef;
  my $station_info = <F>;
  close F;

  # Remove comment lines.
  $station_info =~ s/^#.*\n//mg;

  my @station_blocks = split /\s*\n\s*\n\s*/, $station_info;
  my $hashref;
  my $undefined_code = 'undefined0000';
  foreach (@station_blocks){
    s/\n\s+/$;/g;        # Join continuation lines.
    my @lines = split /\n/;
    my $h;
    my $ordinal = 0;
    foreach (@lines){
      s/\s*$//;			# Remove trailing space.
      my ($key,$val) = split /\s*:\s*/, $_, 2;
      defined $key or next;
      if(!defined $val){
        warn "The $key field has no value for station $h->{code}\n";
        next;
      }
      $key = lc $key;		# One way of implementing case-insensitivity.
      $val =~ s/$;/\n/g;	# Put back any line feeds that were removed.
      $h->{$key} = $val;
    }
    my $c = $h->{code} || $undefined_code++;
    delete $h->{code};
    $hashref->{$c} = $h;
  }
  $hashref;
}
